/* scheduler.cc
 * Copyright (C) 2001 The University of Newcastle upon Tyne
 *
 * Written by Paul Coates (Paul.Coates@ncl.ac.uk)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <arpa/inet.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <netdb.h>
#include "scheduler.h"
#include "parser.h"
#include "discover.h"
#include "process.h"
#include "task.h"
#include "main.h"
#include "session.h"
#include "data.h"
#include "misc.h"
#include "gui.h"

Scheduler::Scheduler(Main *m)
{
	main = m;

	discover = new Process();
	discover->create(1);
	alive = new Process();
	alive->create(2);

	tasks.empty();

	ping_interval = 120;
	topology_interval = 1800;
	autosave_interval = 1800;
	ping_timer.now();			ping_timer.ts += ping_interval;
	topology_timer.now();	topology_timer.ts += topology_interval;
	autosave_timer.now();	autosave_timer.ts += autosave_interval;
}

Scheduler::~Scheduler()
{
	Link *ln;
	Task *t;

	for (ln=tasks.get_head(); ln!=NULL; ln=ln->next)
	{ t = (Task*) ln->item; if (t!=NULL) delete t; }
	tasks.empty();

	alive->type = 0;
	alive->send("REQUEST");
	discover->type = 0;
	discover->send("REQUEST");
}

void Scheduler::addTask(char *str, int type, int lvl, int hop)
{
	Task *t;
	Link *ln;
	int match;
	char buf[64];

	// check for duplicate tasks first
	match = 0;
	for (ln=tasks.get_head(); ln!=NULL; ln=ln->next)
	{
		t = (Task*) ln->item;
		if ((t->type==1) && (type==1) && (strcmp(str, t->cmnd)==0))
		{
			if (t->priority < lvl)
			{
				t->priority = lvl;
				main->gui->discover->update_task_list();
			}
			match = 1;
		}else
		if ((t->type==2) && (type==2))
		{
			// we already have a ping task
			match = 1;
		}
	}

	if (match==0)
	{
		t = new Task();
		strcpy(t->cmnd, str);
		t->priority = lvl;
		t->type = type;
		t->hop = hop;
		tasks.tag(t);
//printf("NEW TASK %s\n", str);
		sprintf(buf, "Discovered device %s", t->cmnd);
		main->gui->discover->discover_message(buf);
		main->gui->discover->update_task_list();
	}
}

void Scheduler::delTasks(int type)
{
	Link *ln;
	Task *t;

	for (ln=tasks.get_head(); ln!=NULL; ln=ln->next)
	{
		t = (Task*) ln->item;

		if (t->type == type)
		{
			delete t;
			ln->item = NULL;
		}
	}
	tasks.tidy();
}

void Scheduler::returnedTask(Process *p)
{
	char buf[BUFSIZE];
	Link *ln2;
	Task *t;

	if (p->get(buf))
	{
		for (ln2=tasks.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			t = (Task*) ln2->item;
			if (t->pid == p->pid)
			{
				if (p->type == 1)	// topology
				{
					if (t->priority>1)
					{
						// topology from discover means no placeholder device
						// so lets create one here
						main->parser->create_placeholder(p->param);
					}
					main->parser->parse(p->param, t->hop);
					// write devices file to include new device
					main->session->save_device_file();
				}else
				if (p->type == 2) // alive
				{
					// do ping stuff
					pingcode();
				}
				delete t;
				ln2->item = NULL;
			}
		}
		tasks.tidy();
		main->gui->discover->update_task_list();
		p->tid = 0;
		strcpy(p->param, "");
	}
}

void Scheduler::assignTask(Process *p)
{
	Link *ln2;
	Task *t;
	char buf[256];

	if (p->tid==0)
	{
		ln2=tasks.get_head();
		while(ln2!=NULL)
		{
			t = (Task*) ln2->item;
			if ((t->pid==0) && (t->type==p->type))
			{
				t->pid = p->pid;
				p->tid = 1;
				strcpy(p->param, t->cmnd);
				p->send("REQUEST");
				if ((t->type==1) && (t->priority>1))
				{
					sprintf(buf, "Assigning task %s", t->cmnd);
					main->gui->discover->discover_message(buf);
				}
				ln2=NULL;
			}else
			{
				ln2=ln2->next;
			}
		}
	}
}

void Scheduler::timeout()
{
	time_t now = time(NULL);

	// 1) check for processes ending and set appropriate flags
	returnedTask(alive);
	returnedTask(discover);

	// 2) check for new tasks

	if (ping_timer.ts <= now)
	{
		if (main->session->dirty=='y') main->session->save_device_file();
		main->scheduler->addTask("", 2, 0, 0);
		ping_timer.ts = now + ping_interval;
	}
	if (topology_timer.ts <= now)
	{
		// it's time to create a set of topology tasks
	}
	if (autosave_timer.ts <= now)
	{
		// if autosave set then save all files
	}

	// 3) assign tasks to processes
	assignTask(alive);
	assignTask(discover);
}

/////////////////////////////////////////////////////////////////////////
// PING CODE
/////////////////////////////////////////////////////////////////////////

void Scheduler::pingcode()
{
	FILE *f;
	char buf[256];
	char buf2[256];
	Device *d;
	Link *ln;
	float ts;
	IPAddress tmpip;

	if (main->session->data==NULL) return;

	for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		d->status = '0';
	}

	sprintf(buf, "%s/ping", main->session->datadir);
	f = fopen(buf, "r");
	if (!f) return;

	for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		fseek(f, 0L, SEEK_SET);
		fscanf(f, "%[^\n]\n", buf); // the date stamp

		while(fscanf(f, "%[^\n]\n", buf)==1)
		{
			if (sscanf(buf, "%[^ ] (%f ms)", buf2, &ts)==2)
			{
				tmpip.setstring(buf2);
				if (tmpip.compare(d->pingname)==0)
					d->status = '1';
			}
		}
	}
	fclose(f);
	main->gui->update_views();
}

